@@ -1,4 +1,6 @@
-#!/usr/bin/python
+#!/usr/bin/env python
+# -*- coding: utf-8 -*-
+# vim: set ts=2 sw=2 et sts=2 ai:
 #
 # Copyright 2009 Google Inc.
 #
@@ -14,10 +16,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-#
+
 # Disable the invalid name warning as we are inheriting from a standard library
 # object.
-# pylint: disable-msg=C6409,W0212
+# pylint: disable=invalid-name,protected-access
 
 """A version of the datetime module which *cares* about timezones.
 
@@ -38,24 +40,35 @@ import datetime
 import os
 import os.path
 import re
+import sys
 import time
 import warnings
 import dateutil.parser
 import dateutil.relativedelta
 import dateutil.tz
 import pytz
-import pytz_abbr
 
+from . import pytz_abbr  # pylint: disable=g-bad-import-order
+
+if sys.platform == "win32":
+  # pylint: disable=g-import-not-at-top
+  from .detect_windows import _detect_timezone_windows
 
 try:
-  # pylint: disable-msg=C6204
+  basestring
+except NameError:
+  # pylint: disable=redefined-builtin
+  basestring = str
+
+try:
+  # pylint: disable=g-import-not-at-top
   import functools
-except ImportError, e:
+except ImportError as e:
 
   class functools(object):
     """Fake replacement for a full functools."""
 
-    # pylint: disable-msg=W0613
+    # pylint: disable=unused-argument
     @staticmethod
     def wraps(f, *args, **kw):
       return f
@@ -84,6 +97,7 @@ def _tzinfome(tzinfo):
   if not isinstance(tzinfo, datetime.tzinfo):
     try:
       tzinfo = pytz.timezone(tzinfo)
+      assert tzinfo.zone in pytz.all_timezones
     except AttributeError:
       raise pytz.UnknownTimeZoneError("Unknown timezone! %s" % tzinfo)
   return tzinfo
@@ -93,31 +107,85 @@ def _tzinfome(tzinfo):
 _localtz = None
 
 
+def localize(dt, force_to_local=True):
+  """Localize a datetime to the local timezone.
+
+  If dt is naive, returns the same datetime with the local timezone, otherwise
+  uses astimezone to convert.
+
+  Args:
+    dt: datetime object.
+    force_to_local: Force all results to be in local time.
+
+  Returns:
+    A datetime_tz object.
+  """
+  if not isinstance(dt, datetime_tz):
+    if not dt.tzinfo:
+      return datetime_tz(dt, tzinfo=localtz())
+    dt = datetime_tz(dt)
+  if force_to_local:
+    return dt.astimezone(localtz())
+  return dt
+
+
+def get_naive(dt):
+  """Gets a naive datetime from a datetime.
+
+  datetime_tz objects can't just have tzinfo replaced with None, you need to
+  call asdatetime.
+
+  Args:
+    dt: datetime object.
+
+  Returns:
+    datetime object without any timezone information.
+  """
+  if not dt.tzinfo:
+    return dt
+  if hasattr(dt, "asdatetime"):
+    return dt.asdatetime()
+  return dt.replace(tzinfo=None)
+
+
 def localtz():
   """Get the local timezone.
 
   Returns:
     The localtime timezone as a tzinfo object.
   """
-  # pylint: disable-msg=W0603
+  # pylint: disable=global-statement
   global _localtz
   if _localtz is None:
     _localtz = detect_timezone()
   return _localtz
 
 
+def localtz_name():
+  """Returns the name of the local timezone."""
+  return str(localtz())
+
+
 def localtz_set(timezone):
   """Set the local timezone."""
-  # pylint: disable-msg=W0603
+  # pylint: disable=global-statement
   global _localtz
   _localtz = _tzinfome(timezone)
 
 
+def require_timezone(zone):
+  """Raises an AssertionError if we are not in the correct timezone."""
+  assert localtz().zone == zone, (
+      "Please set your local timezone to %(zone)s (either in the machine,"
+      "or on Linux by exporting TZ=%(zone)s") % {"zone": zone}
+
+
 def detect_timezone():
   """Try and detect the timezone that Python is currently running in.
 
   We have a bunch of different methods for trying to figure this out (listed in
   order they are attempted).
+    * In windows, use win32timezone.TimeZoneInfo.local()
     * Try TZ environment variable.
     * Try and find /etc/timezone file (with timezone name).
     * Try and find /etc/localtime file (with timezone data).
@@ -129,6 +197,11 @@ def detect_timezone():
   Raises:
     pytz.UnknownTimeZoneError: If it was unable to detect a timezone.
   """
+  if sys.platform == "win32":
+    tz = _detect_timezone_windows()
+    if tz is not None:
+      return tz
+
   # First we try the TZ variable
   tz = _detect_timezone_environ()
   if tz is not None:
@@ -169,27 +242,59 @@ def _detect_timezone_environ():
 def _detect_timezone_etc_timezone():
   if os.path.exists("/etc/timezone"):
     try:
-      tz = file("/etc/timezone").read().strip()
+      tz = open("/etc/timezone").read().strip()
       try:
         return pytz.timezone(tz)
-      except (IOError, pytz.UnknownTimeZoneError), ei:
+      except (IOError, pytz.UnknownTimeZoneError) as ei:
         warnings.warn("Your /etc/timezone file references a timezone (%r) that"
                       " is not valid (%r)." % (tz, ei))
 
     # Problem reading the /etc/timezone file
-    except IOError, eo:
+    except IOError as eo:
       warnings.warn("Could not access your /etc/timezone file: %s" % eo)
 
 
+def _load_local_tzinfo():
+  """Load zoneinfo from local disk."""
+  tzdir = os.environ.get("TZDIR", "/usr/share/zoneinfo/posix")
+
+  localtzdata = {}
+  for dirpath, _, filenames in os.walk(tzdir):
+    for filename in filenames:
+      filepath = os.path.join(dirpath, filename)
+      name = os.path.relpath(filepath, tzdir)
+
+      f = open(filepath, "rb")
+      tzinfo = pytz.tzfile.build_tzinfo(name, f)
+      f.close()
+      localtzdata[name] = tzinfo
+
+  return localtzdata
+
+
 def _detect_timezone_etc_localtime():
+  """Detect timezone based on /etc/localtime file."""
   matches = []
   if os.path.exists("/etc/localtime"):
-    localtime = pytz.tzfile.build_tzinfo("/etc/localtime",
-                                         file("/etc/localtime"))
+    f = open("/etc/localtime", "rb")
+    localtime = pytz.tzfile.build_tzinfo("/etc/localtime", f)
+    f.close()
+
+    # We want to match against the local database because /etc/localtime will
+    # be copied from that. Once we have found a name for /etc/localtime, we can
+    # use the name to get the "same" timezone from the inbuilt pytz database.
+
+    tzdatabase = _load_local_tzinfo()
+    if tzdatabase:
+      tznames = tzdatabase.keys()
+      tzvalues = tzdatabase.__getitem__
+    else:
+      tznames = pytz.all_timezones
+      tzvalues = _tzinfome
 
     # See if we can find a "Human Name" for this..
-    for tzname in pytz.all_timezones:
-      tz = _tzinfome(tzname)
+    for tzname in tznames:
+      tz = tzvalues(tzname)
 
       if dir(tz) != dir(localtime):
         continue
@@ -209,24 +314,29 @@ def _detect_timezone_etc_localtime():
       # We get here iff break didn't happen, i.e. no meaningful attributes
       # differ between tz and localtime
       else:
-        matches.append(tzname)
+        # Try and get a timezone from pytz which has the same name as the zone
+        # which matches in the local database.
+        if tzname not in pytz.all_timezones:
+          warnings.warn("Skipping %s because not in pytz database." % tzname)
+          continue
+
+        matches.append(_tzinfome(tzname))
+
+    matches.sort(key=lambda x: x.zone)
 
     if len(matches) == 1:
-      return _tzinfome(matches[0])
-    else:
-      # Warn the person about this!
-      warning = "Could not get a human name for your timezone: "
-      if len(matches) > 1:
-        warning += ("We detected multiple matches for your /etc/localtime. "
-                    "(Matches where %s)" % matches)
-        return _tzinfome(matches[0])
-      else:
-        warning += "We detected no matches for your /etc/localtime."
-      warnings.warn(warning)
+      return matches[0]
 
-      # Register /etc/localtime as the timezone loaded.
-      pytz._tzinfo_cache['/etc/localtime'] = localtime
-      return localtime
+    if len(matches) > 1:
+      warnings.warn("We detected multiple matches for your /etc/localtime. "
+                    "(Matches where %s)" % matches)
+      return matches[0]
+    else:
+      warnings.warn("We detected no matches for your /etc/localtime.")
+
+    # Register /etc/localtime as the timezone loaded.
+    pytz._tzinfo_cache["/etc/localtime"] = localtime
+    return localtime
 
 
 def _detect_timezone_php():
@@ -246,16 +356,52 @@ def _detect_timezone_php():
       if tomatch == (tz._tzname, -tz._utcoffset.seconds, indst):
         matches.append(tzname)
 
-    # pylint: disable-msg=W0704
+    # pylint: disable=pointless-except
     except AttributeError:
       pass
 
   if len(matches) > 1:
     warnings.warn("We detected multiple matches for the timezone, choosing "
                   "the first %s. (Matches where %s)" % (matches[0], matches))
+  if matches:
     return pytz.timezone(matches[0])
 
 
+class _default_tzinfos(object):
+  """Change tzinfos argument in dateutil.parser.parse() to use pytz.timezone.
+
+  For more details, please see:
+  http://labix.org/python-dateutil#head-c0e81a473b647dfa787dc11e8c69557ec2c3ecd2
+  Usage example:
+
+  >>> dateutil.parser.parse(
+  ...   "Thu Sep 25 10:36:28 UTC 2003",
+  ...   tzinfos=datetime_tz._default_tzinfos())
+  """
+
+  _marker = object()
+
+  def __getitem__(self, key, default=_marker):
+    try:
+      return pytz.timezone(key)
+    except KeyError:
+      if default is self._marker:
+        raise KeyError(key)
+      return default
+
+  get = __getitem__
+
+  def has_key(self, key):
+    return key in pytz.all_timezones
+
+  def __iter__(self):
+    for i in pytz.all_timezones:
+      yield i
+
+  def keys(self):
+    return pytz.all_timezones
+
+
 class datetime_tz(datetime.datetime):
   """An extension of the inbuilt datetime adding more functionality.
 
@@ -319,6 +465,14 @@ class datetime_tz(datetime.datetime):
     obj.is_dst = obj.dst() != datetime.timedelta(0)
     return obj
 
+  def __copy__(self):
+    return type(self)(self)
+
+  def __deepcopy__(self, memo):
+    dpcpy = type(self)(self)
+    memo[id(self)] = dpcpy
+    return dpcpy
+
   def asdatetime(self, naive=True):
     """Return this datetime_tz as a datetime object.
 
@@ -367,9 +521,9 @@ class datetime_tz(datetime.datetime):
     tzinfo = _tzinfome(tzinfo)
 
     d = self.asdatetime(naive=False).astimezone(tzinfo)
-    return datetime_tz(d)
+    return type(self)(d)
 
-  # pylint: disable-msg=C6113
+  # pylint: disable=g-doc-args
   def replace(self, **kw):
     """Return datetime with new specified fields given as arguments.
 
@@ -390,6 +544,11 @@ class datetime_tz(datetime.datetime):
     if "tzinfo" in kw:
       if kw["tzinfo"] is None:
         raise TypeError("Can not remove the timezone use asdatetime()")
+      else:
+        tzinfo = kw["tzinfo"]
+        del kw["tzinfo"]
+    else:
+      tzinfo = None
 
     is_dst = None
     if "is_dst" in kw:
@@ -401,9 +560,10 @@ class datetime_tz(datetime.datetime):
 
     replaced = self.asdatetime().replace(**kw)
 
-    return datetime_tz(replaced, tzinfo=self.tzinfo.zone, is_dst=is_dst)
+    return type(self)(
+        replaced, tzinfo=tzinfo or self.tzinfo.zone, is_dst=is_dst)
 
-  # pylint: disable-msg=C6310
+  # pylint: disable=line-to-long
   @classmethod
   def smartparse(cls, toparse, tzinfo=None):
     """Method which uses dateutil.parse and extras to try and parse the string.
@@ -414,12 +574,12 @@ class datetime_tz(datetime.datetime):
     Other valid formats include:
       "now" or "today"
       "yesterday"
-      "tommorrow"
+      "tomorrow"
       "5 minutes ago"
       "10 hours ago"
       "10h5m ago"
       "start of yesterday"
-      "end of tommorrow"
+      "end of tomorrow"
       "end of 3rd of March"
 
     Args:
@@ -471,7 +631,9 @@ class datetime_tz(datetime.datetime):
     elif toparselower == "yesterday":
       dt -= datetime.timedelta(days=1)
 
-    elif toparselower == "tommorrow":
+    elif toparselower in ("tomorrow", "tommorrow"):
+      # tommorrow is spelled wrong, but code out there might be depending on it
+      # working
       dt += datetime.timedelta(days=1)
 
     elif "ago" in toparselower:
@@ -563,14 +725,30 @@ class datetime_tz(datetime.datetime):
       tzinfo = localtz()
     return obj.astimezone(tzinfo)
 
+  # pylint: disable=redefined-outer-name
+  @classmethod
+  def combine(cls, date, time, tzinfo=None):
+    """date, time, [tz] -> datetime with same date and time fields."""
+    if tzinfo is None:
+      tzinfo = localtz()
+    return cls(datetime.datetime.combine(date, time), tzinfo)
+
   today = now
 
+  # pylint: disable=unused-argument
   @staticmethod
   def fromordinal(ordinal):
     raise SyntaxError("Not enough information to create a datetime_tz object "
                       "from an ordinal. Please use datetime.date.fromordinal")
 
 
+# We can't use datetime's absolute min/max otherwise astimezone will fail.
+datetime_tz.min = datetime_tz(
+    datetime.datetime.min+datetime.timedelta(days=2), pytz.utc)
+datetime_tz.max = datetime_tz(
+    datetime.datetime.max-datetime.timedelta(days=2), pytz.utc)
+
+
 class iterate(object):
   """Helpful iterators for working with datetime_tz objects."""
 
@@ -687,24 +865,27 @@ def _wrap_method(name):
 
   # Have to give the second argument as method has no __module__ option.
   @functools.wraps(method, ("__name__", "__doc__"), ())
-  def wrapper(*args, **kw):
-    r = method(*args, **kw)
+  def wrapper(self, *args, **kw):
+    r = method(self, *args, **kw)
 
-    if isinstance(r, datetime.datetime) and not isinstance(r, datetime_tz):
-      r = datetime_tz(r)
+    if isinstance(r, datetime.datetime) and not isinstance(r, type(self)):
+      r = type(self)(r)
     return r
 
   setattr(datetime_tz, name, wrapper)
 
-for methodname in ["__add__", "__radd__", "__rsub__", "__sub__", "combine"]:
+for methodname in ["__add__", "__radd__", "__rsub__", "__sub__"]:
 
   # Make sure we have not already got an override for this method
   assert methodname not in datetime_tz.__dict__
+  # pypy 1.5.0 lacks __rsub__
+  if hasattr(datetime.datetime, methodname):
+    _wrap_method(methodname)
 
-  _wrap_method(methodname)
+__all__ = [
+    "datetime_tz", "detect_timezone", "iterate", "localtz",
+    "localtz_set", "timedelta", "_detect_timezone_environ",
+    "_detect_timezone_etc_localtime", "_detect_timezone_etc_timezone",
+    "_detect_timezone_php", "localize", "get_naive", "localtz_name",
+    "require_timezone"]
 
-
-__all__ = ['datetime_tz', 'detect_timezone', 'iterate', 'localtz',
-    'localtz_set', 'timedelta', '_detect_timezone_environ',
-    '_detect_timezone_etc_localtime', '_detect_timezone_etc_timezone',
-    '_detect_timezone_php']
